// ----------------------------------------------------------------------------
// Store.java
// Persistent store of declarations and proofs.
// $Id: Store.java,v 1.13 2006/01/15 15:30:39 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Communication;

import java.io.*;
import java.util.*;
import java.util.zip.*;

import nl.tue.win.riaca.openmath.lang.OMObject;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import fmrisc.External.GeneralHashFunctionLibrary;
import fmrisc.External.NewOMDOMReader;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.ProofNavigator.*;
import fmrisc.Proving.*;

/**
 * Persistent store of declarations and proofs.
 * 
 * For a declaration <name> of <kind> (where <kind> in {type,value,formula}) 
 * the following files are created:
 * 
 * <kind>_<name>_decl.xgz
 *   Gzipped XML file holding the OMDoc representation of the declaration.
 *   
 * <kind>_<name>_hash.txt (not for proofs)
 *   A single text line with a hash value for the representation (a 
 *   non-negative integer number in decimal representation).
 *   
 * <kind>_<name>_time.txt
 *   A single text line with a time stamp (a non-negative integer number
 *   in decimal representation).
 *   
 * <kind>_<name>_refs.xml
 *   XML file representing the symbols referenced in the declaration:
 *   <fmrisc:refs>
 *     <fmrisc:<kind> time="timestampvalue"><name></fmrisc:<kind>>
 *     ...
 *   </fmrisc:refs>
 * 
 * Writing to the Store:
 * 
 * If an entity is to be written to the store, it is assumed that all the 
 * symbols it depends on have already been updated in the store.
 * 
 * If the entity is not yet contained in the store, we write it to the 
 * store and are done.
 * 
 * Otherwise, we check whether the
 * symbols listed in the corresponding "refs" file are stored with the
 * same time stamps as listed in the file. If yes, then the hash value
 * of the internal representation is computed. If this value is the same
 * as the hash value of the "hash" file, we assume that the definition
 * has not changed and we are done without updating the entity. 
 * Otherwise (i.e., if one of the previous conditions is not satisfied), 
 * we write the entity to the store updating all files.
 *  
 * Currently, we do not read declaration contents from the store.
 * 
 * For a proof, the following files are created:
 * 
 * proof_<name>_decl.xgz
 *   Gzipped XML file representing the proof of the formula <name>.
 *   <fmrisc:proof name="formulaname" closed="true|false">
 *     <fmrisc:state>
 *       <fmrisc:command> </fmrisc:command>
 *       <fmrisc:state>...</fmrisc:state>
 *       ...
 *     </fmrisc:state>
 *   </fmrisc:proof>
 *   
 * <kind>_<name>_refs.xml
 *   XML file representing the symbols referenced in the declaration:
 * 
 * Writing a proof proceeds as explained for the declarations above.
 * 
 * When a proof is read from the store, it is assumed that all the 
 * symbols in memory have already been updated in the store.
 * 
 * The proof is returned with status "skeleton=yes" and the status
 * "closed" as listed in the file.
 * 
 * The proof's "trusted" value is set to "true" iff the symbols listed 
 * in the "refs" file are stored with the same time stamps as listed in 
 * the file. Replaying such a proof must not cause errors.
 * 
 * The proof's "absolute" value is set to "true" iff 
 * the proof is "trusted" and "closed" and all formulas
 * listed in the "refs" file have absolute proofs.
 *****************************************************************************/
public final class Store
{
  private File directory;
  
  /***************************************************************************
   * Create instance of store for denoted directory.
   * @param directory for storing the declared objects
   **************************************************************************/
  public Store(File directory)
  {
    this.directory = directory;
  }
  
  /****************************************************************************
   * Construct file from name.
   * @param name of file
   * @return the file
   ***************************************************************************/
  private File getFile(String name)
  {
    return new File(directory, name);
  }
  
  /***************************************************************************
   * Write proof to store updating the dependencies file.
   * @param proof the proof
   * @return true iff no problem occured in storing
   **************************************************************************/
  public boolean write(Proof proof)
  {
    // writing the proof also constructs collection of referenced symbols 
    boolean okay = writeProof(proof);
    if (!okay) return false;
    
    // write referenced symbols
    File dfile = getFile(getProofDepName(proof.getSymbol()));
    okay = writeDep(dfile, proof.getReferenced());
    return okay;
  }
  
  /***************************************************************************
   * Write proof to store.
   * @param proof the proof
   * @return true iff no problem occured in storing
   **************************************************************************/
  private boolean writeProof(Proof proof)
  {
    String name = getProofDeclName(proof.getSymbol());
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument();
      Node node = proof.toNode(document);
      document.appendChild(node);
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      File file = getFile(name);
      OutputStream out = 
        new GZIPOutputStream (new BufferedOutputStream(new FileOutputStream(file)));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
      return true;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
  }
  
  /***************************************************************************
   * Check whether store holds proof of formula.
   * @param symbol a formula symbol.
   * @return true iff store holds a proof of the formula.
   **************************************************************************/
  public boolean existsProof(FormulaSymbol symbol)
  {
    File file = getFile(getProofDeclName(symbol));
    return file.exists();
  }
  
  /***************************************************************************
   * Read proof for formula from store.
   * @param symbol a formula symbol.
   * @return the proof of the formula (or null, if an error occurred).
   **************************************************************************/
  public Proof readProof(FormulaSymbol symbol)
  {
    try
    {
      File file = getFile(getProofDeclName(symbol));
      InputStream in = 
        new GZIPInputStream(new BufferedInputStream(new FileInputStream(file)));
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(in);
      if (document == null) return null;
      Node root = document.getFirstChild();
      if (root == null) return null;
      removeWhitespace(root);
      Proof proof = Proof.toProof(symbol, root);
      if (proof == null) return null;
      File dfile = getFile(getProofDepName(symbol));
      if (dfile.isFile())
      {
        HashMap dep = readDep(dfile);
        if (dep == null) return null;
        proof.setTrusted(!hasChanged(dep));
        proof.setAbsolute(true); // symbol itself occurs in dependence map
        proof.setAbsolute(absoluteOnly(dep));
      }
      else
      {
        proof.setTrusted(false);
        proof.setAbsolute(false);
      }
      return proof;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Write value declaration to store.
   * @param decl the declaration
   * @param referenced the symbols referenced in the definition of the symbol
   * @return true iff no problem occured in storing
   **************************************************************************/
  public boolean write(Declaration decl, Collection referenced)
  {
    File file = getFile(getDepName(decl.getSymbol()));
    if (hasChanged(decl))
    {
      boolean okay = write(decl);
      if (!okay) return false;
      okay = writeDep(file, referenced);
      if (!okay) return false;
    }
    else
    {
      if (!file.isFile()) 
      {
        State.setError("file " + file + " not found");
        return false;
      }
      HashMap dep = readDep(file);
      if (dep == null) return false;
      if (hasChanged(dep, referenced))
      {
        boolean okay = writeDep(file, referenced);
        if (!okay) return false;
      }
      else 
      {
        if (dep.size() != referenced.size())
        {
          boolean okay = writeDep(file, referenced);
          if (!okay) return false;
        }
        return true;
      }
    }
    boolean okay = writeHash(decl);
    if (!okay) return false;
    okay = writeStamp(decl.getSymbol());
    return okay;
  }
  
  /***************************************************************************
   * Signal whether declaration has changed compared to that in store.
   * @param decl a declaration
   * @return true iff store declaration is different from current one
   **************************************************************************/
  public boolean hasChanged(Declaration decl)
  {
    File file = getFile(getHashName(decl.getSymbol()));
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(file));
      long hash0 = Long.parseLong(reader.readLine());
      reader.close();
      long hash1 = getHash(decl);
      return hash0 != hash1;
    }
    catch(FileNotFoundException e)
    {
      // file not found, must assume that content has changed
      return true;
    }
    catch(IOException e)
    {
      // cannot read file, must assume that content has changed
      return true;
    }
    catch(NumberFormatException e)
    {
      // invalid hash code, must assume that content has changed
      return true;
    }
  }
  
  /***************************************************************************
   * Signal whether dependencies have changed compared to that in store.
   * @param dep a dependencies map.
   * @param referenced collection of symbols of referenced values
   * @return true iff stored dependencies different from current one
   **************************************************************************/
  public boolean hasChanged(HashMap dep, Collection referenced)
  {
    Iterator iterator = referenced.iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      Long stamp0 = (Long)dep.get(symbol);
      if (stamp0 == null) return true;
      long stamp1 = getStamp(symbol);
      if (stamp0.longValue() != stamp1) return true;
    }
    return false;
  }
  
  /***************************************************************************
   * Signal whether dependencies have changed time stamps
   * compared to that listed in dependency map.
   * @param dep map of dependency mappings.
   * @return true iff time stamps have changed.
   **************************************************************************/
  private boolean hasChanged(HashMap dep)
  {
    Iterator iterator = dep.keySet().iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      Long stamp0 = (Long)dep.get(symbol);
      if (stamp0 == null) return true;
      long stamp1 = getStamp(symbol);
      if (stamp0.longValue() != stamp1) return true;
    }
    return false;
  }
   
  /***************************************************************************
   * Signal whether all formulas in dependency map have absolute proofs.
   * @param dep map of dependencies.
   * @return true iff all formulas in map have absolute proofs.
   **************************************************************************/
  private boolean absoluteOnly(HashMap dep)
  {
    Iterator iterator = dep.keySet().iterator();
    while (iterator.hasNext())
    {
      Symbol symbol = (Symbol)(iterator.next());
      if (!(symbol instanceof FormulaSymbol)) continue;
      FormulaSymbol fsymbol = (FormulaSymbol)symbol;
      if (fsymbol.isAxiom()) continue;
      Proof proof = fsymbol.getProof();
      if (proof == null) return false;
      if (!proof.isAbsolute()) return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Read dependencies from file.
   * @param file a file holding the dependencies.
   * @return map of Symbol->Long pairs representing the dependencies
   *         (null, if some error occured)
   **************************************************************************/
  private HashMap readDep(File file)
  {
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(file);
      if (document == null) return null;
      Node root = document.getFirstChild();
      if (root == null) return null;
      removeWhitespace(root);
      String name = root.getLocalName();
      if (!name.equals("refs"))
      {
        State.setError("root node in file " + file.getAbsolutePath() + 
          " has name " + name + " (\"refs\" expected)");
        return null;
      }
      HashMap result = new HashMap();
      NodeList list = root.getChildNodes();
      int n = list.getLength();
      for (int i=0; i<n; i++)
      {
        Node node = list.item(i);
        name = node.getLocalName();
        NamedNodeMap attributes = node.getAttributes();
        long stamp = -1;
        if (attributes != null)
        {
          Node attribute = attributes.getNamedItem("time");
          if (attribute != null)
            stamp = Long.parseLong(((Attr)attribute).getValue());
        }
        if (stamp == -1)
        {
          State.setError("Invalid symbol number " + i + " in file " 
              + file.getAbsolutePath() + " (no valid \"time\" attribute):\n" + node.toString());
          return null;
        }
        Node child = node.getFirstChild();
        if (child == null || !(child instanceof Text))
        {
          State.setError("Invalid symbol number " + i + " in file " 
            + file.getAbsolutePath() + " (no text):\n" + node.toString());
          return null;
        } 
        Text text = (Text)child;
        Identifier ident = Construct.reference(text.getData());
        Environment env = State.getEnvironment();
        Symbol csymbol = null;
        if (name.equals("type"))
          csymbol = env.getTypeSymbol(ident);
        else if (name.equals("value"))
          csymbol = env.getValueSymbol(ident);
        else if (name.equals("formula"))
          csymbol = env.getFormulaSymbol(ident);
        else
          State.setError("Invalid symbol number " + i + " in file " 
              + file.getAbsolutePath() + " (unknown element):\n"
              + node.toString());
        if (csymbol == null) return null;
        result.put(csymbol, new Long(stamp));
      }  
      return result;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Write dependencies to file.
   * @param file the file to write to.
   * @param referenced collection of symbols of referenced values
   * @return true iff no problem occured in writing.
   **************************************************************************/
  public boolean writeDep(File file, Collection referenced)
  {
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    try
    {
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument();
      Element root = document.createElementNS(
        State.NamespaceURI, State.NamespacePrefix + ":refs");
      document.appendChild(root);
      root.setAttribute("xmlns:" + State.NamespacePrefix, State.NamespaceURI);
      Iterator iterator = referenced.iterator();
      while (iterator.hasNext())
      {
        Symbol symbol = (Symbol)iterator.next();
        Identifier ident = symbol.getIdentifier();
        String kind = null;
        if (symbol instanceof TypeSymbol)
          kind = "type";
        else if (symbol instanceof ValueSymbol)
          kind = "value";
        else if (symbol instanceof FormulaSymbol)
          kind = "formula";
        Element ref = document.createElementNS(
          State.NamespaceURI, State.NamespacePrefix + ":" + kind);
        root.appendChild(ref);
        ref.setAttribute("time", Long.toString(getStamp(symbol)));
        Text text = document.createTextNode(ident.getName());
        ref.appendChild(text);
      }
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      StreamResult result = new StreamResult(file);
      transformer.transform(source, result);
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(NumberFormatException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
 
  /****************************************************************************
   * write time stamp of symbol
   * @param symbol a symbol
   **************************************************************************/
  private boolean writeStamp(Symbol symbol)
  {
    String name = getStampName(symbol);
    File file = getFile(name);
    try
    {
      PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      writer.println(new Date().getTime());
      writer.close();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * get time stamp of declaration (-1, if an error occured)
   * @decl a declaration
   **************************************************************************/
  private long getStamp(Symbol symbol)
  {
    String name = getStampName(symbol);
    File file = getFile(name);
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(file));
      long time = Long.parseLong(reader.readLine());
      reader.close();
      return time;
    }
    catch(FileNotFoundException e)
    {
      return -1;
    }
    catch(IOException e)
    {
      return -1;
    }
    catch(NumberFormatException e)
    {
      return -1;
    }
  }
  
  /****************************************************************************
   * Compute hash code for declaration.
   * @param decl the declaration.
   * @return its hash code.
   ***************************************************************************/
  private long getHash(Declaration decl)
  {
    String rep = ASTUtil.canonicalText(decl);
    return GeneralHashFunctionLibrary.RSHash(rep);
  }
  
  /****************************************************************************
   * write hash code of symbol
   * @decl a symbol
   **************************************************************************/
  private boolean writeHash(Declaration decl)
  {
    String name = getHashName(decl.getSymbol());
    File file = getFile(name);
    try
    {
      PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      writer.println(getHash(decl));
      writer.close();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Write declaration to store.
   * @param decl the declaration
   * @return true iff no problem occured in storing
   **************************************************************************/
  private boolean write(Declaration decl)
  {
    String name = getDeclName(decl.getSymbol());
    File file = getFile(name);
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.newDocument(); 
      OMDoc omdoc = new OMDoc(document);
      Element node = omdoc.toNode(decl);
      document.appendChild(node);
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      DOMSource source = new DOMSource(document);
      OutputStream out = 
        new GZIPOutputStream(
          new BufferedOutputStream(new FileOutputStream(file)));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
    }
    catch(IOException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(TransformerConfigurationException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    catch(TransformerException e)
    {
      State.setError("exception" + e.toString());
      return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Read declaration from file.
   * @param file path to a file holding the declaration.
   * @return the declaration (null, if an error occured).
   **************************************************************************/
  public static Declaration readDeclaration(File file)
  {
    try
    {
      InputStream in = 
        new GZIPInputStream(new BufferedInputStream(new FileInputStream(file)));
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      Document document = builder.parse(in);
      if (document == null) return null;
      Node root = document.getFirstChild();
      if (root == null) return null;
      removeWhitespace(root);
      return toDeclaration(root);
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
    catch(SAXException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Get file base name for symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getBaseName(Symbol symbol)
  {
    return getPrefix(symbol) + "_" + symbol.getIdentifier().getName();
  }
  
 
  /***************************************************************************
   * Get file name for content of symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getDeclName(Symbol symbol)
  {
    return getBaseName(symbol) + "_decl.xgz";
  }
  
  /***************************************************************************
   * Get file name for time stamp of symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getStampName(Symbol symbol)
  {
    return getBaseName(symbol) + "_time.txt";
  }
  
  /***************************************************************************
   * Get file name for hash code of symbol
   * @param symbol a symbol
   **************************************************************************/
  private static String getHashName(Symbol symbol)
  {
    return getBaseName(symbol) + "_hash.txt";
  }
  
  /***************************************************************************
   * Get dependencies file name for symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getDepName(Symbol symbol)
  {
    return getBaseName(symbol) + "_refs.xml";
  }
  
  /***************************************************************************
   * Return unique prefix for symbol kind.
   * @return a string denoting the symbol kind
   **************************************************************************/
  private static String getPrefix(Symbol symbol)
  {
    if (symbol instanceof TypeSymbol)
      return "type";
    else if (symbol instanceof ValueSymbol)
      return "value";
    else if (symbol instanceof FormulaSymbol)
      return "formula";
    return null;
  }
  
  /***************************************************************************
   * Get proof base name for formula symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getProofBaseName(FormulaSymbol symbol)
  {
    return "proof_" + symbol.getIdentifier().getName();
  }
  
  /***************************************************************************
   * Get proof file name for content of symbol.
   * @param symbol a symbol.
   **************************************************************************/
  private static String getProofDeclName(FormulaSymbol symbol)
  {
    return getProofBaseName(symbol) + "_decl.xgz";
  }
  
  /***************************************************************************
   * Get proof dependencies file name for symbol.
   * @param symbol a symbol
   **************************************************************************/
  private static String getProofDepName(FormulaSymbol symbol)
  {
    return getProofBaseName(symbol) + "_refs.xml";
  }
  
  /***************************************************************************
   * Convert DOM node to OpenMath object.
   * @param node the DOM node holding the OpenMath representation.
   * @return the corresponding OpenMath node.
   **************************************************************************/
  private static OMObject toOpenMath(Node node)
  {
    NewOMDOMReader reader = new NewOMDOMReader(node);
    try
    {
      return reader.readObject();
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    }
  }
  
  /***************************************************************************
   * Convert DOM node to expression.
   * @param node the DOM representation of the expression.
   * @return the expression (null, if an error occured).
   **************************************************************************/
  public static Expression toExpression(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toExpression(om);
  }
  
  /***************************************************************************
   * Convert DOM node to type.
   * @param node the DOM representation of the type.
   * @return the type (null, if an error occured).
   **************************************************************************/
  public static Type toType(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toType(om);
  }
  
  /***************************************************************************
   * Convert DOM node to identifier.
   * @param node the DOM representation of the identifier.
   * @return the type (null, if an error occured).
   **************************************************************************/
  public static Identifier toIdentifier(Node node)
  {
    OMObject om = toOpenMath(node);
    if (om == null) return null;
    return OpenMath.toIdentifier(om);
  }
  
  /***************************************************************************
   * Convert DOM node to declaration.
   * @param node the DOM representation of the declaration.
   * @return the declaration (null, if an error occured).
   **************************************************************************/
  public static Declaration toDeclaration(Node node)
  {
    return OMDoc.toDeclaration(node);
  }
  
  /***************************************************************************
   * Removes all text nodes that only contain whitespace from DOM tree.
   * @param node the root of the tree that is to be cleaned from whitespace.
   **************************************************************************/
  public static void removeWhitespace(Node node)
  {
    Vector empty = new Vector();
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    for (int i=0; i<n; i++)
    {
      Node child = children.item(i);
      if (!(child instanceof Text))
      {
        removeWhitespace(child);
        continue;
      }
      String data = ((Text)child).getData();
      boolean iswhite = true;
      int l = data.length();
      for (int j=0; j<l; j++)
      {
        char ch = data.charAt(j);
        if (!Character.isWhitespace(ch))
        {
          iswhite = false;
          break;
        }
      }
      if (iswhite) empty.add(child);
    }
    Iterator iterator = empty.iterator();
    while (iterator.hasNext())
      node.removeChild((Node)iterator.next());
  }      
}
