// ----------------------------------------------------------------------------
// Flatten.java
// The "flatten" command
// $Id: Flatten.java,v 1.16 2006/03/29 15:49:47 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.Proving.*;
import fmrisc.Syntax.*;

import java.io.PrintWriter;
import java.util.*;

import org.w3c.dom.Node;

/**
 * 
 * Command "flatten": decompose compound formulas without splitting the state.
 *
 */
public final class Flatten extends ProofCommandBase
{
  public final static String Name = "flatten";
  
  /***************************************************************************
   * create a "flatten" command
   **************************************************************************/
  public Flatten()
  {
    super(Name);
  }
  
  /****************************************************************************
   * process command "flatten": 
   * decompose compound formulas without splitting the state.
   * Applies the following rules: 
   * - A1 /\ A2, ... |- B, ... => A1, A2, ... |- B
   * - A, ... |- B1 \/ B2, ... => A, ..., ~B1 |- B2
   * - A, ... |- B1 => B2, ... => A, ..., B1 |- B2
   * - ~A, ... |- B => negate(A), ... |- B
   * - A, ... |- ~B => A, ... |- negate(B)
   * - A1 <=> A2, ... |- B, ... => A1=>A2 /\ A2=>A1, ... |- B
   * - A, ... |- B1 <=> B2, ... => A, ... |- B1=>B2 /\ B2=>B1
   * - A xor A2, ... |- B, ... => A1=>~A2 /\ ~A1=>A2
   * - A, ... |- B1 xor B2, ... => A, ... |- B1=>~B2 /\ ~B1=>B2
   * where "negate(F)" pushes negation inside the formula as far as possible.
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  { 
    ProofState original= state;
    boolean changedOnce = false;
    boolean changed;
    do
    {
      changed = false;
      ProofState state0 = new ProofState(state, state.getEnvironment());
      
      Formula[] assumptions = state.getAssumptions();
      int n = assumptions.length;
      Vector assumptions0 = new Vector();
      for (int i=0; i<n; i++)
      {
        Formula assumption = assumptions[i];
        Expression formula = assumption.getExpression();
        if (formula instanceof AndFormula)
        {
          AndFormula f = (AndFormula)formula;
          assumptions0.add(new Formula(state0, f.getFirst(), false));
          assumptions0.add(new Formula(state0, f.getSecond(), false));
          changed = true;
          continue;
        }
        if (formula instanceof NotFormula)
        {
          NotFormula f = (NotFormula)formula;
          Formula f0 = new Formula(state0, ProofUtil.negate(f.getBase()), false);
          if (assumption.getLabel().equals(f0.getLabel()))
            assumptions0.add(assumption);
          else
          {
            assumptions0.add(f0);
            changed = true;
          }
          continue;
        }
        if (formula instanceof EquivalentFormula)
        {
          EquivalentFormula f = (EquivalentFormula)formula;
          Expression f0 = f.getFirst();
          Expression f1 = f.getSecond();
          AndFormula c = Construct.andFormula(
              Construct.impliesFormula(f0.instantiate(), f1.instantiate()),
              Construct.impliesFormula(f1.instantiate(), f0.instantiate()));         
          assumptions0.add(new Formula(state0, c, false));
          changed = true;
          continue;
        }
        if (formula instanceof NotEquivalentFormula)
        {
          NotEquivalentFormula f = (NotEquivalentFormula)formula;
          Expression f0 = f.getFirst();
          Expression f1 = f.getSecond();
          AndFormula c = Construct.andFormula(
            Construct.impliesFormula(f0.instantiate(), 
              Construct.notFormula(f1.instantiate())),
            Construct.impliesFormula(Construct.notFormula(f0.instantiate()), 
              f1.instantiate()));
          assumptions0.add(new Formula(state0, c, false));
          changed = true;
          continue;
        }
        assumptions0.add(new Formula(state0, assumption));
      }
      
      Formula[] goals = state.getGoals();
      int m = goals.length;
      Vector goals0 = new Vector();
      for (int i=0; i<m; i++)
      {
        Formula goal = goals[i];
        Expression formula = goal.getExpression();
        if (formula instanceof OrFormula)
        {
          OrFormula f = (OrFormula)formula;
          assumptions0.add(new Formula(state0, 
            ProofUtil.negate(f.getFirst()), false));
          goals0.add(new Formula(state0, f.getSecond(), true));
          changed = true;
          continue;
        }
        if (formula instanceof ImpliesFormula)
        {
          ImpliesFormula f = (ImpliesFormula)formula;
          assumptions0.add(new Formula(state0, f.getFirst(), false));
          goals0.add(new Formula(state0, f.getSecond(), true));
          changed = true;
          continue;
        }
        if (formula instanceof NotFormula)
        {
          NotFormula f = (NotFormula)formula;
          Formula f0 = new Formula(state0, ProofUtil.negate(f.getBase()), true);
          if (goal.getLabel().equals(f0.getLabel()))
            goals0.add(goal);
          else
          {
            goals0.add(f0);
            changed = true;
          }
          continue;
        }
        if (formula instanceof EquivalentFormula)
        {
          EquivalentFormula f = (EquivalentFormula)formula;
          Expression f0 = f.getFirst();
          Expression f1 = f.getSecond();
          AndFormula c = Construct.andFormula(
            Construct.impliesFormula(f0.instantiate(), f1.instantiate()),
            Construct.impliesFormula(f1.instantiate(), f0.instantiate()));
          goals0.add(new Formula(state0, c, true));
          changed = true;
          continue;
        }
        if (formula instanceof NotEquivalentFormula)
        {
          NotEquivalentFormula f = (NotEquivalentFormula)formula;
          Expression f0 = f.getFirst();
          Expression f1 = f.getSecond();
          AndFormula c = Construct.andFormula(
            Construct.impliesFormula(f0.instantiate(), 
              Construct.notFormula(f1.instantiate())),
            Construct.impliesFormula(Construct.notFormula(f0.instantiate()), 
              f1.instantiate()));
          goals0.add(new Formula(state0, c, true));
          changed = true;
          continue;
        }
        goals0.add(new Formula(state0, goal));
      }
      
      state0.setAssumptions(Formula.toArray(assumptions0));
      state0.setGoals(Formula.toArray(goals0));
      state = state0;
      changedOnce = changedOnce || changed;
    }
    while (false && changed);
    
    if (!changedOnce) return null;
    state.setParent(original);
    return new ProofState[] { state };
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    return new Flatten();
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
  }
}
