// ----------------------------------------------------------------------------
// Formula.java
// A formula occuring in a proof state.
// $Id: Formula.java,v 1.12 2006/03/28 16:11:29 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving;

import fmrisc.Syntax.*;

import java.io.*;
import java.util.Vector;
/**
 * A formula occuring in a proof state.
 */
public final class Formula
{
  private ProofState state;   // the state in which the formula occurs
  private Expression formula; // the formula expression
  private boolean isGoal;     // true if formula is goal (assumption, otherwise)
  private String label;       // label for unique formula reference within state
  
  private final static int len=3; // length of label in number of characters
  private static int counter=0;   // counter to make labels unique
  
  /****************************************************************************
   * construct a formula from an expression
   * @param state the proof state in which the formula occurs
   * @param formula the expression representing the formula
   * @param isGoal true if formula is a goal, false, if it is an assumption.
   **************************************************************************/
  public Formula(ProofState state, Expression formula, boolean isGoal)
  {
    this.state = state;
    this.isGoal = isGoal; // polarity may be used by hash label
    setExpression(formula);
  }
  
  /****************************************************************************
   * construct a copy of the given formula in the denoted state
   * @param state the proof state in which the new formula occurs
   * @param formula a given formula
   **************************************************************************/
  public Formula(ProofState state, Formula formula)
  {
    this.state = state;
    this.isGoal = formula.isGoal(); // polarity may be used by hash label
    setExpression(formula.getExpression());
  }
   
  /***************************************************************************
   * hash formula to short string.
   * @return a string representing the formula uniquely in the current state
   **************************************************************************/
  private String hash()
  {
    String rep = formula.toString();
    String label = ProofUtil.hash(rep, len);
    Formula formula0 = state.getFormula(label);
    if (formula0 != null && !rep.equals(formula0.getExpression().toString()))
    {
      // resolve the unlikely case of a hash collision
      label = label + "_" + counter;
      counter = counter+1;
    }
    return label;
  }
  
  /****************************************************************************
   * Get proof state of formula.
   * @return the proof state to which the formula belongs.
   **************************************************************************/
  public ProofState getState()
  {
    return state;
  }
  
  /****************************************************************************
   * Get formula expression.
   * @return the expression of the formula
   **************************************************************************/
  public Expression getExpression()
  {
    return formula;
  }
  
  /****************************************************************************
   * Set formula expression.
   * @param formula the formula expression
   **************************************************************************/
  public void setExpression(Expression formula)
  {
    this.formula = formula;
    this.label = hash();
  }
  
  /****************************************************************************
   * get formula label
   * @return a label by which the formula can be uniquely referenced within
   *         the current proof state
   **************************************************************************/
  public String getLabel()
  {
    return label;
  }
  
  /****************************************************************************
   * get formula isGoal
   * @return false if formula is assumption, true if it is goal
   **************************************************************************/
  public boolean isGoal()
  {
    return isGoal;
  }

  /***************************************************************************
   * print formula on output stream as "[label] formula"
   * @param out the stream on which to print the formula
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.print("[" + label + "] ");
    formula.print(out);
  }
  
  /***************************************************************************
   * converts vector v of Formula objects to array
   * @param v a vector holding Formula objects
   * @return an array holding these objects
   **************************************************************************/
  public static Formula[] toArray(Vector v)
  {
    return (Formula[])v.toArray(new Formula[v.size()]);
  }
}
